# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)


class AccuracyError(Exception):
    """Is raised if the result of a computation is inaccurate."""


class InversionError(Exception):
    pass
